/*
 * simple stat routines
 *          by k-chinen@is.aist-nara.ac.jp, 1999
 *
 * $Id$
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>
#include <limits.h>
#include <sys/types.h>
#ifdef HAVE_ALLOCA_H
#include <alloca.h>
#endif

#include <errno.h>

#include "sim_hist.h"
#include "mem.h"
#include "msg.h"

int
sim_hist_clear(sim_hist_rec *srec)
{
    int i;
    int chk;

    chk = MUTEX_LOCK(&srec->busy);

    for(i=0;i<srec->max;i++) {
        srec->hcount[i] = 0L;
    }
    srec->count = 0L;
    srec->igcount = 0L;
#ifdef LONG_MAX
    srec->v_max = LONG_MIN;
    srec->v_min = LONG_MAX;
#else
    srec->v_max = -2147483646;
    srec->v_min = 2147483647;
#endif

    chk = MUTEX_UNLOCK(&srec->busy);

    return 0;
}

int
sim_hist_alloc_new(sim_hist_rec *addr, char *name, int sz, long *ar)
{
    sim_hist_rec *ret;
    int chk;

    if(addr==NULL) {
        return -1;
    }
    ret = addr;

    chk = pthread_mutex_init(&addr->busy, NULL);

    chk = MUTEX_LOCK(&addr->busy);
    strcpy(ret->name, name);
    ret->max = sz;
    ret->hcount = ar;
    chk = MUTEX_UNLOCK(&addr->busy);

    sim_hist_clear(ret);

    return 0;
}

sim_hist_rec*
sim_hist_new(char *name, int sz)
{
    int chk;
    sim_hist_rec *ret;
    int i;

    ret = (sim_hist_rec*) malloc(sizeof(sim_hist_rec));
    if(ret==NULL) {
        return NULL;
    }
    MLOCK((void*)ret, sizeof(sim_hist_rec));

    chk = pthread_mutex_init(&ret->busy, NULL);

    chk = MUTEX_LOCK(&ret->busy);
    strcpy(ret->name, name);
    ret->max = sz;
    ret->hcount = (long*) malloc(sizeof(long)*sz);
    if(ret->hcount==NULL) {
        ret = NULL;
    }
    for(i=0;i<ret->max;i++) {
        ret->hcount[i] = 0L;
    }
    MLOCK((void*)ret->hcount, sizeof(long)*sz);
    chk = MUTEX_UNLOCK(&ret->busy);

    sim_hist_clear(ret);

    return ret;
}


int
sim_hist_reset(sim_hist_rec *rec)
{
    int i;
    int chk;

    chk = MUTEX_LOCK(&rec->busy);

    rec->vsum = 0L;
    rec->count = 0L;
    for(i=0;i<rec->max;i++) {
        rec->hcount[i] = 0L;
    }
    rec->igcount = 0L;
    chk = MUTEX_UNLOCK(&rec->busy);

    return 0;
}

int
sim_hist_add_value(sim_hist_rec *rec, long cv)
{
    int chk;

    chk = MUTEX_LOCK(&rec->busy);

    rec->hcount[rec->count%rec->max] = cv;
    rec->count++;
    if(cv>rec->v_max) {
        rec->v_max = cv;
    }
    if(cv<rec->v_min) {
        rec->v_min = cv;
    }

    chk = MUTEX_UNLOCK(&rec->busy);

    return 0;
}

int
sim_hist_average(sim_hist_rec *rec)
{
    int i;
    int chk;
    long sum;
    long av;

    chk = MUTEX_LOCK(&rec->busy);

    sum = 0L;
    for(i=0;i<rec->max;i++) {
        sum += rec->hcount[i];
    }

    if(rec->count==0) {
        av = -1;
    }
    else
    if(rec->count<rec->max) {
        av = sum/rec->count;
    }
    else {
        av = sum/rec->max;
    }

    chk = MUTEX_UNLOCK(&rec->busy);

    return av;
}

double
sim_hist_average_d(sim_hist_rec *rec)
{
    int i;
    int chk;
    long sum;
    double av;

    chk = MUTEX_LOCK(&rec->busy);

    sum = 0L;
    for(i=0;i<rec->max;i++) {
        sum += rec->hcount[i];
    }

    if(rec->count==0) {
        av = -1.0;
    }
    else
    if(rec->count<rec->max) {
        av = ((double)sum)/rec->count;
    }
    else {
        av = ((double)sum)/rec->max;
    }

    chk = MUTEX_UNLOCK(&rec->busy);

    return av;
}

static
int
comp(const void *x, const void *y)
{
    return *((int*)x) - *((int*)y);
}

int
sim_hist_percentail(sim_hist_rec *rec, int ratio)
{
    int i;
    int chk;
    int av;
    int pos;
    long *buf;
    int n;

    chk = MUTEX_LOCK(&rec->busy);

    buf = (long*)alloca(sizeof(long)*rec->max);
    if(buf==NULL) {
        av = -1;
        goto out;
    }

    if(rec->max>rec->count) {
        n = rec->count;
    }
    else {
        n = rec->max;
    }
Trace("max %d, count %d -> n %d\n", rec->max, rec->count, n);

    memcpy(buf, &rec->hcount[0], sizeof(long)*n);
    qsort(buf, n, sizeof(long), comp);

for(i=0;i<n;i++) {
    Trace(" %4d: %9d\n", i, buf[i]);
}

    pos = (int)(((double)n*ratio)/100.0);
Trace("pos %d -> %9d\n", pos, buf[pos]);

    av = buf[pos];

out:
    chk = MUTEX_UNLOCK(&rec->busy);

    return av;
}


int
sim_hist_show(sim_hist_rec *rec, FILE *fp)
{
    int i;
    int chk;
    long sum;
    long av;

    chk = MUTEX_LOCK(&rec->busy);

    fprintf(fp, "=== %s ===\n", rec->name);
    fprintf(fp, "count %ld, sz %ld, max %ld, min %ld\n",
        rec->count, rec->max, rec->v_max, rec->v_min);

    sum = 0L;
    for(i=0;i<rec->max;i++) {
        fprintf(fp, "%3d %c %d\n",
            i, i==rec->count%rec->max ? '*' : ':', rec->hcount[i]);
        sum += rec->hcount[i];
    }

    if(rec->count==0) {
        av = -1;
    }
    else
    if(rec->count<rec->max) {
        av = sum/rec->count;
    }
    else {
        av = sum/rec->max;
    }
    fprintf(fp, "running average %ld (sum %ld)\n", av, sum);

    chk = MUTEX_UNLOCK(&rec->busy);
}


#if 0

int
sim_hist_show(sim_hist_rec *rec, FILE *fp)
{
    int i;
    int chk;
    long tsum, asum;

    chk = MUTEX_LOCK(&rec->busy);

    fprintf(fp, "=== %s ===\n", rec->name);

    if(rec->count<=0) {
        fprintf(fp, "no sample\n");
        chk = MUTEX_UNLOCK(&rec->busy);
        return -1;
    }


    fprintf(fp, "access %d time, sum %d -> ave %.3f\n",
        rec->count, rec->vsum,
        rec->count<=0 ? -1.0 : (double)rec->vsum/rec->count);

    tsum = 0L;
    for(i=0;i<rec->max;i++) {
        tsum += rec->hcount[i];
    }
    tsum += rec->igcount;
    asum = tsum;

    fprintf(fp, "#%6s %9s %6s %9s %6s\n",
            "value", "count", "count%", "CDF", "CDF%");
    tsum = 0L;
    for(i=0;i<rec->max;i++) {
        tsum += rec->hcount[i];
        if(rec->hcount[i]>0) {
            fprintf(fp, " %6d %9d %6.2f %9d %6.2f\n",
                i, rec->hcount[i],
                asum<=0 ? -1.0 : 100.0*(double)rec->hcount[i]/asum,
                tsum,
                asum<=0 ? -1.0 : 100.0*(double)tsum/asum);
        }
    }

    tsum += rec->igcount;
    fprintf(fp, " %6s %9d %6.2f %9d %6.2f\n",
        "others", rec->igcount,
        asum<=0 ? -1.0 : 100.0*(double)rec->igcount/asum,
        tsum,
        asum<=0 ? -1.0 : 100.0*(double)tsum/asum);


    fprintf(fp, "\n");

    chk = MUTEX_UNLOCK(&rec->busy);

    return 0;
}

int
sim_hist_show_lock(sim_hist_rec *rec, FILE *fp)
{
    int i;
    int chk;
    long tsum, asum;


    fprintf(fp, "=== %s ===\n", rec->name);

    if(rec->count<=0) {
        fprintf(fp, "no sample\n");
        return -1;
    }


    fprintf(fp, "access %d time, sum %d -> ave %.3f\n",
        rec->count, rec->vsum,
        rec->count<=0 ?  (double)-1.0 : (double)rec->vsum/rec->count);

    tsum = 0L;
    for(i=0;i<rec->max;i++) {
        tsum += rec->hcount[i];
    }
    tsum += rec->igcount;
    asum = tsum;

    fprintf(fp, "#%6s %9s %7s %9s %7s\n",
            "value", "count", "count%", "CDF", "CDF%");
    tsum = 0L;
    for(i=0;i<rec->max;i++) {
        tsum += rec->hcount[i];
        if(rec->hcount[i]>0) {
            fprintf(fp, " %6d %9d %6.2f %9d %6.2f\n",
                i, rec->hcount[i], 
                asum<=0 ?  (double)-1.0 : 100.0*(double)rec->hcount[i]/asum,
                tsum,
                asum<=0 ?  (double)-1.0 : 100.0*(double)tsum/asum);
        }
    }

    tsum += rec->igcount;
    fprintf(fp, " %6s %9d %6.2f %9d %6.2f\n",
        "others", rec->igcount, 
        asum<=0 ? (double)-1.0 : 100.0*(double)rec->igcount/asum,
        tsum,
        asum<=0 ? (double)-1.0 : 100.0*(double)tsum/asum);


    fprintf(fp, "\n");

    return 0;
}

int
sim_hist_show_lock_fd(sim_hist_rec *rec, int ofd)
{
    int i;
    int chk;
    long tsum, asum;


    fdprintf(ofd, "=== %s === (without lock)\n", rec->name);

    if(rec->count<=0) {
        fdprintf(ofd, "no sample\n");
        return -1;
    }


    fdprintf(ofd, "access %d time, sum %d -> ave %.3f\n",
        rec->count, rec->vsum,
        rec->count<=0 ? (double)-1.0 : (double)rec->vsum/rec->count);

    tsum = 0L;
    for(i=0;i<rec->max;i++) {
        tsum += rec->hcount[i];
    }
    tsum += rec->igcount;
    asum = tsum;

    fdprintf(ofd, "#%6s %9s %7s %9s %7s\n",
            "value", "count", "count%", "CDF", "CDF%");
    tsum = 0L;
    for(i=0;i<rec->max;i++) {
        tsum += rec->hcount[i];
        if(rec->hcount[i]>0) {
            fdprintf(ofd, " %6d %9d %6.2f %9d %6.2f\n",
                i, rec->hcount[i],
                asum<=0 ? (double)-1.0 : 100.0*(double)rec->hcount[i]/asum,
                tsum,
                asum<=0 ? (double)-1.0 : 100.0*(double)tsum/asum);
        }
    }

    tsum += rec->igcount;
    fdprintf(ofd, "#%6s %9d %6.2f %9d %6.2f\n",
        "others", rec->igcount, 
        asum<=0 ? (double)-1.0 : 100.0*(double)rec->igcount/asum,
        tsum,
        asum<=0 ? (double)-1.0 : 100.0*(double)tsum/asum);


    fdprintf(ofd, "\n");

    return 0;
}

int
sim_hist_show_HTML_nolock_fd(sim_hist_rec *rec, int ofd)
{
    int i;
    int chk;
    long tsum, asum;
    int v;


    fdprintf(ofd, "<table border=\"0\" cellspacing=\"0\" cellpadding=\"2\">\n\
<tr bgcolor=\"#ff9999\"> <th colspan=\"7\">%s (without lock)\n", rec->name);


    if(rec->count<=0) {
        fdprintf(ofd, "<tr><th>no sample\n");
        goto out;
    }


    tsum = 0L;
    for(i=0;i<rec->max;i++) {
        tsum += rec->hcount[i];
    }
    tsum += rec->igcount;
    asum = tsum;

    fdprintf(ofd, "<tr bgcolor=\"#ffcccc\">\n\
<th>%s <th>%s <th>%s <th>%s <th>%s <th width=\"390\">%s <th width=\"390\">%s\n",
            "value", "count", "PDF%", "sum", "CDF%",
            "graph of PDF", "graph of CDF");
    tsum = 0L;
    for(i=0;i<rec->max;i++) {
        tsum += rec->hcount[i];
        if(rec->hcount[i]>0) {
            fdprintf(ofd,
"<tr><td align=\"right\"> %d \
<td align=\"right\"> %d \
<td align=\"right\"> %6.2f \
<td align=\"right\"> %d \
<td align=\"right\"> %6.2f  ",
                i, rec->hcount[i],
                asum<=0 ? -1.0 : 100.0*(double)rec->hcount[i]/asum,
                tsum,
                asum<=0 ? -1.0 : 100.0*(double)tsum/asum);

            if(asum<=0) {
                v = 0;
            }
            else {
                v = (int)(100.0*(double)rec->hcount[i]/asum);
            }

            if(v==100) {
                fdprintf(ofd,
"<td width=\"390\"><img width=\"200\" src=\"/img/bw200p100.jpg\"> 100  ");
            }
            else
            if(v<=0) {
                fdprintf(ofd,
"<td width=\"390\"> 0  ");
            }
            else
            if(v<=10 || v%10==0) {
                fdprintf(ofd,
"<td width=\"390\"><img width=\"%d\" src=\"/img/bw200p%02d.jpg\"> %d  ",
                    v*2, v, v);
            }
            else {
                fdprintf(ofd,
"<td width=\"390\">\
<img width=\"%d\" src=\"/img/bw200p%02d.jpg\">\
<img width=\"%d\" src=\"/img/bw200p%02d.jpg\"> %d  ",
                    (v/10)*20, (v/10)*10, (v%10)*2, v%10 , v);
            }


            if(asum<=0) {
                v = 0;
            }
            else {
                v = (int)(100.0*(double)tsum/asum);
            }

            if(v==100) {
                fdprintf(ofd,
"<td width=\"390\"><img width=\"200\" src=\"/img/bw200p100.jpg\"> 100  ");
            }
            else
            if(v<=0) {
                fdprintf(ofd,
"<td width=\"390\"> 0  ");
            }
            else
            if(v<=10 || v%10==0) {
                fdprintf(ofd,
"<td width=\"390\"><img width=\"%d\" src=\"/img/bw200p%02d.jpg\"> %d  ",
                    v*2, v, v);
            }
            else {
                fdprintf(ofd,
"<td width=\"390\"><img width=\"%d\" src=\"/img/bw200p%02d.jpg\">\
<img width=\"%d\" src=\"/img/bw200p%02d.jpg\"> %d  ",
                    (v/10)*20, (v/10)*10, (v%10)*2, v%10 , v);
            }

            fdprintf(ofd, "\n");
        }
    }

    if(rec->igcount>0) {
        tsum += rec->igcount;
        fdprintf(ofd,
"<tr><td align=\"center\"> %s\n\
<td align=\"right\"> %d\n\
<td align=\"right\"> %6.2f\n\
<td align=\"right\"> %d\n\
<td align=\"right\"> %6.2f\n",
            "others", rec->igcount,
            asum<=0 ? -1.0 : 100.0*(double)rec->igcount/asum,
            tsum, 
            asum<=0 ? -1.0 : 100.0*(double)tsum/asum);

        if(asum<=0) 
            v = 0;
        else
            v = (int)(100.0*(double)rec->igcount/asum);

        if(v==100) {
            fdprintf(ofd, "<td><img width=\"200\" src=\"/img/bw200p100.jpg\"> 100  ");
        }
        else
        if(v<=0) {
            fdprintf(ofd, "<td> 0  ");
        }
        else
        if(v<=10 || v%10==0) {
            fdprintf(ofd, "<td><img width=\"%d\" src=\"/img/bw200p%02d.jpg\"> %d  ",
                v*2, v, v);
        }
        else {
            fdprintf(ofd,
"<td><img width=\"%d\" src=\"/img/bw200p%02d.jpg\"><img width=\"%d\" src=\"/img/bw200p%02d.jpg\"> %d  ",
                (v/10)*20, (v/10)*10, (v%10)*2, v%10, v);
        }

        if(asum<=0)
            v = 0;
        elese
            v = (int)(100.0*(double)tsum/asum);
        if(v==100) {
            fdprintf(ofd, "<td><img width=\"200\" src=\"/img/bw200p100.jpg\"> 100  ");
        }
        else
        if(v<=0) {
            fdprintf(ofd, "<td> 0  ");
        }
        else
        if(v<=10 || v%10==0) {
            fdprintf(ofd, "<td><img width=\"%d\" src=\"/img/bw200p%02d.jpg\"> %d  ",
                v*2, v, v);
        }
        else {
            fdprintf(ofd,
"<td><img width=\"%d\" src=\"/img/bw200p%02d.jpg\"><img width=\"%d\" src=\"/img/bw200p%02d.jpg\"> %d  ",
                (v/10)*20, (v/10)*10, (v%10)*2, v%10, v);
        }

        fdprintf(ofd, "\n");

    }


    fdprintf(ofd,
"<tr><th bgcolor=\"#ffcccc\">total <td align=\"right\"> %d\n\
<tr><th bgcolor=\"#ffcccc\">average <td align=\"right\"> %d\n\
<tr><th bgcolor=\"#ffcccc\">average <td align=\"right\"> %.1f\n",
        rec->count, rec->vsum, 
        rec->count<=0 ? -1.0 : (double)rec->vsum/rec->count);

out:
    fdprintf(ofd, "</table>\n");

    return 0;
}

int
sim_hist_show_skip_lock_fd(sim_hist_rec *rec, int ofd, int sk)
{
    int i;
    int chk;
    long tsum, asum;


    fdprintf(ofd, "=== %s === (with skip without lock)\n", rec->name);

    if(rec->count<=0) {
        fdprintf(ofd, "no sample\n");
        return -1;
    }


    fdprintf(ofd, "access %d time, sum %d -> ave %.3f\n",
        rec->count, rec->vsum,
        rec->count<=0 ? -1.0 : (double)rec->vsum/rec->count);

    tsum = 0L;
    for(i=0;i<rec->max;i++) {
        tsum += rec->hcount[i];
    }
    tsum += rec->igcount;
    asum = tsum;

    fdprintf(ofd, "#%6s %9s %7s %9s %7s\n",
            "value", "count", "count%", "CDF", "CDF%");
    tsum = 0L;
    for(i=0;i<rec->max;i++) {
        tsum += rec->hcount[i];
#if 0
        if(rec->hcount[i]>0) {
            fdprintf(ofd, " %6d %9d %6.2f %9d %6.2f\n",
                i, rec->hcount[i], 100.0*(double)rec->hcount[i]/asum,
                tsum, 100.0*(double)tsum/asum);
        }
#else
        if(i%sk==0 && tsum>0) {
            fdprintf(ofd, " %6d %9d %6.2f %9d %6.2f\n",
                i, rec->hcount[i],
                asum<=0 ? -1.0 : 100.0*(double)rec->hcount[i]/asum,
                tsum, 
                asum<=0 ? -1.0 : 100.0*(double)tsum/asum);
        }
#endif
    }

    tsum += rec->igcount;
    fdprintf(ofd, "#%6s %9d %6.2f %9d %6.2f\n",
        "others", rec->igcount,
        asum<=0 ? -1.0 : 100.0*(double)rec->igcount/asum,
        tsum,   
        asum<=0 ? -1.0 : 100.0*(double)tsum/asum);


    fdprintf(ofd, "\n");

    return 0;
}

int
sim_hist_show_short(sim_hist_rec *rec, FILE *fp)
{
    int chk;

    chk = MUTEX_LOCK(&rec->busy);
    fprintf(fp, "access %d time, sum %d -> ave %.3f\n",
        rec->count, rec->vsum,
        rec->count<=0 ? -1.0 : (double)rec->vsum/rec->count);
    chk = MUTEX_UNLOCK(&rec->busy);

    return 0;
}


#endif
